#!/usr/bin/perl
#
# Search the directory for a UID supplied on the command line.  The
# script requires a keytab and will create the ticket cache from
# the keytab.

use Stanford::Directory;
use strict;

##############################################################################
# Helper routines
##############################################################################

# Print the usage message and die
sub print_usage {
    die "Usage: $0 <keytab file> <principal> <someuid>\n";
}

# Handle command line arguments
sub get_arguments {

    # Die if nothing on the command line
    if (!$ARGV[0]) {
        print_usage;
    }

    # Make sure the kerberos ticket file exists
    my $krb_file = $ARGV[0];
    if (!$krb_file) {
        print "Keytab file '$krb_file' not found\n";
        print_usage;
    }

    # Separate the principal service/instance into it parts
    if (!$ARGV[1]) {
        print "Kerberos principal missing\n";
        print_usage;
    }

    my $krb_princ = $ARGV[1];
    my $krb_service;
    my $krb_instance;
    if ($krb_princ =~ q{(.*)/(.*)}) {
        $krb_service  = $1;
        $krb_instance = $2;
    } else {
        print "Invalid Kerberos principal $krb_princ\n";
        print_usage;
    }

    # Lastly pull the uid from the command line
    if (!$ARGV[2]) {
        print "UID missing\n";
        print_usage;
    }
    my $uid = $ARGV[2];

    return ($krb_file, $krb_service, $krb_instance, $uid);
}

##############################################################################
# Main Routine
##############################################################################

my ($krb_file, $krb_service, $krb_instance, $uid) = get_arguments();

my $dir = new Stanford::Directory;
$dir->set(
    ldap_server => 'ldap.stanford.edu',
    mechanism   => 'GSSAPI',
    basedn      => 'cn=accounts,dc=stanford,dc=edu',
    keytab      => $krb_file,
    service     => $krb_service,
    instance    => $krb_instance,
    ticket_file => '/tmp/example-keytab-mail-ticket.tgt'
);

my @attrs = ('suSeasSUNetID', 'suMailDrop');
my @entries = $dir->ldap_query("uid=$uid", @attrs);
if (!@entries) {
    if ($dir->error) {
        my $msg = 'Error type:' . $dir->error_type . "\n";
        $msg .= 'Error:' . $dir->error . "\n";
        die $msg;
    } else {
        print "\n";
        print "UID '$uid' not found in the directory\n";
        print "\n";
        print "or\n";
        print "\n";
        print "$krb_service/$krb_instance does not have access to the \n";
        print "UID attribute in the cn=accounts,dc=stanford,dc=edu branch\n";
        print "of the directory.\n";
        print "\n";
    }
} else {
    for my $entry_ref (@entries) {
        my $this_uid  = ($entry_ref->suseassunetid)[0];
        my $mail_drop = ($entry_ref->suMailDrop)[0];
        printf "UID:%s suMailDrop:%s\n", $this_uid, $mail_drop;
    }
}

exit;
