#######################################################################
#
# Directory.pm - Perl module for searching the Stanford Directory
#
#----------------------------------------------------------------------
#  Copyright  2000-2013 Board of Trustees, Leland Stanford Jr. University
#
#  This module is free software; you may redistribute it and/or modify it
#  under the same terms as Perl itself.
#######################################################################

## no critic (Perl::Critic::Policy::CodeLayout::ProhibitParensWithBuiltins);

package Stanford::Directory;

use strict;
use warnings;
use autodie;

use vars qw(@ISA @EXPORT @EXPORT_OK $AUTOLOAD);

use Authen::Krb5;
use Authen::SASL;
use Carp;
use MIME::Base64;
use Socket;
use Stanford::Directory::Entry;
use English;

use Net::LDAP;
use Net::LDAP::Control;
use Net::LDAPapi;

require Exporter;

@ISA = qw(Exporter);

# Control OIDs
my $LDAP_CONTROL_VALSORT = '1.3.6.1.4.1.4203.666.5.14';

# Global SASL context.  This is workaround to overcome initialization
# problems with Net::LDAP, Authen::SASL, and Authen::Krb5 in the same
# context.
my $SASL;

#------------------------------------------------------------------------
# new (constructor) - set up directory connection parameters

sub new {
    my $that = shift;
    my $class = ref($that) || $that;

    my %fields = (
        ldap_conn   => undef,
        ldap_server => 'ldap.stanford.edu',
        mechanism   => undef,
        version     => 3,
        scope       => 'sub',
        service     => undef,
        instance    => undef,
        keytab      => undef,
        basedn      => 'dc=stanford, dc=edu',
        ticket_file => "/tmp/directory.$UID.tkt",
        error       => q{},
        error_type  => q{},
        keytab_init => 0,
        debug       => 0,
        weights     => 0,
        start_tls   => 0,
        net_ldapapi => 0,
    );

    my $self = {
        _permitted => \%fields,
        %fields,
    };
    bless $self, $class;
    return $self;
}

#------------------------------------------------------------------------
# AUTOLOAD - used to get and set parameters individually

sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self);

    my $name = $AUTOLOAD;
    $name =~ s/.*://xms;    # strip fully-qualified portion
    return if $name =~ /DESTROY/xms;    # don't do DESTROY

    if (!exists $self->{_permitted}->{$name}) {
        carp "$type doesn't have a '$name'";
        return;
    }

    if (@_) { $self->{$name} = shift }
    return $self->{$name};
}

#------------------------------------------------------------------------
# set - set parameters en mass

sub set {
    my $self = shift;
    my $type = ref($self);
    my $name;
    while ($name = shift) {
        if (!exists $self->{_permitted}->{$name}) {
            carp "$type doesn't have a '$name'";
            shift;
            next;
        }
        $self->{$name} = shift;
    }
    return;
}

#------------------------------------------------------------------------
# as_string - just a toy to list a Directory for debugging and such

sub as_string {
    my $self = shift;
    my $type = ref($self) or croak "$self is not an object";
    my $string;

    for my $key (sort keys %{ $self->{_permitted} }) {
        $string .= sprintf("%16s : %s\n", $key, $self->{$key});
    }
    return $string;
}

#------------------------------------------------------------------------
# Initialize SASL

sub sasl_init {
    my $self = shift;
    if (!$SASL) {
        $self->dbg('Creating SASL object with mechanism ' . $self->mechanism());
        $SASL = Authen::SASL->new($self->mechanism());
    }
    return;
}

#------------------------------------------------------------------------
# ldap_query - Using the given filter, contact the LDAP server and see
#              what we get back.  Parse it into a list of Entries, one
#              for each record returned.

sub ldap_query {
    my ($self, $filter, @attrs) = @_;

    # Clear error, error_type
    $self->error(q{});
    $self->error_type(q{});

    # If necessary, convet $filter, to an array ref.
    my @filters = ();
    if (!ref $filter) {
        @filters = ($filter);
    } else {
        @filters = @{ $filter };
    }

    # Connect to the directory
    $self->open_directory();
    if ($self->error) {
        croak($self->error_type . ': ' . $self->error . "\n");
    }

    if ($self->net_ldapapi) {
        return $self->ldap_query_net_ldapapi(\@filters, \@attrs) ;
    } else {
        return $self->ldap_query_net_ldap(\@filters, \@attrs) ;
    }
}

sub ldap_query_net_ldap {
    my ($self, $filters_aref, $attrs_aref) = @_;

    my @filters = @{ $filters_aref };
    my @attrs   = @{ $attrs_aref };

    my $ld = $self->ldap_conn;

    # Create any query controls requested
    my $ctrl;
    if ($self->weights) {
        # Create an ASN.1 boolean and set it to 1
        my $asn = Convert::ASN1->new;
        $asn->prepare('SEQUENCE { b BOOLEAN }');
        my $bool1 = $asn->encode(b => 1);
        # Create the control
        $ctrl = Net::LDAP::Control->new(
            type     => $LDAP_CONTROL_VALSORT,
            critical => 1,
            value    => $bool1
        );
    }

    # Perform the query
    my %entries;
    for my $filter (@filters) {
        my @entries;
        $self->dbg("[net_ldap] Performing search with filter $filter");
        my $status;
        if ($ctrl) {
            $status = $ld->search(
                base    => $self->basedn,
                scope   => $self->scope,
                filter  => $filter,
                attrs   => \@attrs,
                control => $ctrl
            );
        } else {
            $status = $ld->search(
                base   => $self->basedn,
                scope  => $self->scope,
                filter => $filter,
                attrs  => \@attrs
            );
        }

        # Check for errors
        if ($status->is_error) {
            my $err;
            $err .= $status->error_name() . ' - ' . $status->error() . "\n";
            if ($err =~ /exceeded$/xms) {
                $err .= ' (try narrowing your search)';
            }
            $self->error_type('Search');
            $self->error($err);
            if ($err !~ /exceeded$/xms) {
                return;
            }
        }

        # No errors, cool. Grab the entries make them Entries.
        my @entrys;
        @entrys = $status->all_entries;
        for my $ent (@entrys) {
            my $entry = Stanford::Directory::Entry->new();
            $entry->dn($ent->dn);
            for my $attr ($ent->attributes()) {
                my $key = lc($attr);
                $entry->$key($ent->get_value($attr));
            }
            push(@entries, $entry);
        }
        if (scalar(@filters) == 1) {
            return @entries;
        } else {
            $entries{$filter} = \@entries;
        }
    }

    return %entries;
}

sub ldap_query_net_ldapapi {
    my ($self, $filters_aref, $attrs_aref) = @_;

    my @filters = @{ $filters_aref };
    my @attrs   = @{ $attrs_aref };

    my $ld = $self->ldap_conn;

    my $scope;
    if ($self->scope eq 'sub') {
        $scope = LDAP_SCOPE_SUBTREE;
    } elsif ($self->scope eq 'base') {
        $scope = LDAP_SCOPE_BASE;
    } elsif ($self->scope eq 'one') {
        $scope = LDAP_SCOPE_ONELEVEL;
    } else {
        croak 'cannot do an Net::LDAPapi search with scope ' . $self->scope;
    }

    # Perform the query
    my %entries;
    for my $filter (@filters) {
        my @entries;
            $self->dbg('[net_ldapapi] Performing search:');
            $self->dbg('[net_ldapapi]   base:   ' . $self->basedn);
            $self->dbg('[net_ldapapi]   scope:  ' . $scope) ;
            $self->dbg('[net_ldapapi]   filter: ' . $filter) ;
            $self->dbg('[net_ldapapi]   attrs:  ' . join(', ', @attrs)) ;

        my $status = $ld->search_s($self->basedn, $scope, $filter, \@attrs, 0);
        if ($status != LDAP_SUCCESS) {
            $self->dbg('[net_ldapapi] LDAP search was not successful') ;
            my $m = q{};
            if ($ld->errstring) {
                $m .= $ld->errstring;
            }
            if ($ld->extramsg) {
                $m .= ' - ' . $ld->extramsg;
            }
            if (!$m) {
                $m = "Unknown LDAP error status = $status";
            }
            $self->error_type('Search');
            $self->error($m);

            return;
        } else {
            $self->dbg('[net_ldapapi] LDAP search was successful') ;
        }

        # No errors, cool. Grab the entries make them Entries.
        my $entry_ref = $ld->get_all_entries;
        my %entrys    = %{ $entry_ref };
        for my $dn (keys %entrys) {
            my $entry = Stanford::Directory::Entry->new();
            $self->dbg("[net_ldapapi] current dn: $dn") ;
            $entry->dn($dn);
            my $attributes_href = $entrys{$dn};
            for my $attribute (keys %{ $attributes_href }) {
                my $key = lc($attribute);
                my $values_aref = $attributes_href->{$attribute};
                $entry->$key(@{ $values_aref });
            }
            push(@entries, $entry);
        }
        if (scalar(@filters) == 1) {
            return @entries;
        } else {
            $entries{$filter} = \@entries;
        }
    }

    return %entries;
}


#------------------------------------------------------------------------
# This function is here for backwards compatibility and there is no
# reason to call it any more.  This query function was originally
# added as a performance improvement and is not really needed anymore
# since the LDAP connections are kept until directory_close is
# explicitly called.

sub ldap_persistent_query {
    my ($self, $filter, @attrs) = @_;
    my @entries = $self->ldap_query($filter, @attrs);
    return @entries;
}

#------------------------------------------------------------------------
# manage_ksrvtgt - make sure we have a valid ticket (this isn't really
#                  a method, because it might not really belong here).

sub _manage_ksrvtgt {
    my $self = shift;
    my ($user, $instance, $keytab, $file) = @_;

    $self->dbg('Creating ticket cache');

    # Check for keytab
    if (!-r $keytab) {
        return "Cannot read keytab file, $keytab";
    }

    my $status;
    {
        Authen::Krb5::init_context();
        Authen::Krb5::init_ets();
        my $client = Authen::Krb5::parse_name("$user/$instance");
        my $server = Authen::Krb5::parse_name('krbtgt/stanford.edu');
        my $cc     = Authen::Krb5::cc_resolve("FILE:$file");
        $cc->initialize($client);
        my $kt = Authen::Krb5::kt_resolve($keytab);
        Authen::Krb5::get_in_tkt_with_keytab($client, $server, $kt, $cc)
          or croak(Authen::Krb5::error() . " while getting Kerberos ticket\n");
        $self->dbg("ticket cache file: FILE:$file");

    }

    # Set environment variable.
    $ENV{KRB5CCNAME} = "FILE:$file";
    $self->dbg("set KRB5CCNAME to FILE:$file");

    if ($status) {
        return
          "Cannot guarantee ticket for \"$user.$instance\" (status $status)";
    }
    return 1;
}

#------------------------------------------------------------------------
# Open a connection to the directory

sub open_directory {
    my ($self) = @_;

    # We only need to connect once
    return if $self->ldap_conn;

    # Manage the tgt if we've got a keytab --
    # This should be abstracted into something like $self->authenticate
    if ($self->keytab) {
        my $err = $self->_manage_ksrvtgt(
            $self->service, $self->instance,
            $self->keytab,  $self->ticket_file,
        );
        if ($err ne '1') {
            $self->error_type('System');
            $self->error($err);
            return ();
        }
    }

    if ($self->net_ldapapi) {
        return $self->open_directory_net_ldapapi()
    } else {
        return $self->open_directory_net_ldap()
    }
}

sub open_directory_net_ldap {
    my ($self) = @_;

    # Connect to the server
    my $server_name = gethostbyaddr(inet_aton($self->ldap_server()), AF_INET);
    $self->dbg("Connecting to $server_name");
    my $ld = Net::LDAP->new($server_name, version => $self->version());
    if ($ld == -1) {
        $self->error_type('System');
        $self->error('Could not connect to directory server ('
              . $self->ldap_server
              . ')');
        return ();
    }

    if ($self->start_tls) {
        $self->dbg(q[set start_tls to 'require']);
        $ld->start_tls(verify => 'require');
    }

    # Bind to the directory
    my $status;
    if ($self->mechanism) {
        $self->sasl_init();
        $self->dbg('SASL bind to the directory');
        $status = $ld->bind(q{}, sasl => $SASL);
    } else {
        $status = $ld->bind(q{});
    }
    if ($status->code) {
        croak(  'Bind error: ('
              . $status->error_name . ') '
              . $status->error_text);
    }

    $self->ldap_conn($ld);

    return;
}

sub open_directory_net_ldapapi {
    my ($self) = @_;

    # Connect to the server
    my $server_name = gethostbyaddr(inet_aton($self->ldap_server()), AF_INET);
    $self->dbg("Connecting to $server_name");
    my $ld = Net::LDAPapi->new($server_name);
    if ($ld == -1) {
        $self->error_type('System');
        $self->error('Could not connect to directory server ('
              . $self->ldap_server
              . ')');
        return ();
    }

    # Bind to the directory (we only support GSSAPI)
    $ld->sasl_parms(-mech=>'GSSAPI');

    my $status = $ld->bind_s(-type=>LDAP_AUTH_SASL);
    if ($status != LDAP_SUCCESS) {
        $ld->unbind;
        $self->error_type('System');
        $self->error('Could not bind to directory server ('
                     . $self->ldap_server
                     . ')');
        return ();
    }

    $self->ldap_conn($ld);

    return;
}

#------------------------------------------------------------------------
# Close the open directory connection

sub close_directory {
    my ($self) = @_;
    my $ld = $self->ldap_conn;
    if ($ld) {
        $ld->unbind;
    }
    return;
}

#------------------------------------------------------------------------
# Display debugging messages.

sub dbg {
    my ($self, $msg) = @_;

    if ($self->debug) {
        return _dbg($msg);
    }
    return;
}

sub _dbg {
    my ($msg) = @_;
    print {*STDOUT} "DEBUG: $msg\n" or croak("debugging to STDOUT failed: $ERRNO");
    return;
}

1;
__END__

=head1 NAME

Stanford::Directory - Perl module for searching the Stanford Directory

=head1 SYNOPSIS

use Stanford::Directory;

$dir = new Stanford::Directory;

$dir->set(attribute =E<gt> 'value', ...);

@entries = $dir->ldap_query($filter [, @attributes ]);
or
@entries = $dir->ldap_query(\@filter [, @attributes ]);

=head1 PERSISTENT SYNOPSIS

use Stanford::Directory;

$dir = new Stanford::Directory;

$dir->set(attribute =E<gt> 'value', ...);

$dir->open_directory;


@entries = $dir->ldap_persistent_query($filter  [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter2 [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter3 [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter4 [, @attributes ]);
...

$dir->close_directory;

=head1 DESCRIPTION

B<Stanford::Directory> is a light-weight object for accessing the
Stanford Directory.  Use it to maintain the information necessary
to connect to the directory and to search the directory using LDAP
filters.

=head1 METHODS

B<new>

Allocates and returns a new B<Stanford::Directory> object.

B<set(attribute =E<gt> 'value', ...)>

Use I<set> to set the various B<Stanford::Directory> attributes (see
the B<ATTRIBUTES> section below).  Each attribute can also be set or
retrieved by using its name as a method.

B<ldap_query($ldap_filter [, @attributes ])>
B<ldap_query(\@ldap_filter [, @attributes ])>

The I<ldap_query> method makes a directory search using the supplied
I<ldap_filter>.  It returns an array of B<Stanford::Directory::Entry>
objects.  The I<attributes> array contains the names of directory
entry attributes to include in each B<Stanford::Directory::Entry>.  If
it is empty or omitted all of the directory entry attributes are
included.  Alternately, you can provide a reference to an array of
filters to supply to the I<ldap_query> method.  The benefit of using
an array of filters is that multiple searches will be done inside a
single bind to the directory server, which is much speedier than doing
multiple single-filter queries.  It will return a hash of references
to B<Stanford::Directory::Entry> objects that are keyed by search
filter.

B<open_directory>

Opens up a persistent connection to the directory.

B<ldap_persistent_query($ldap_filter [, @attributes])>

The I<ldap_persistent_query> method behaves the same way as the
I<ldap_query> method, except that it does not close the connection
when it finishes the search.  Subsequent searches are done using the
same connection.  If the connection gets closed between searches,
I<ldap_persistent_query> will open a new one, and then resume
searching.

This method is obsolete.  The ldap_query method no longer closes the
connection when it completes.

B<close_directory>
Close a persistent connection to the directory.

=head1 ATTRIBUTES

The B<Stanford::Directory> object has the following attributes (with
default values, if any):

  ldap_server     LDAP server name (ldap.stanford.edu)
  mechanism       undef (for anonymous binds) or GSSAPI
  start_tls       Do a STARTTLS after initial bind (0 or 1)
  service         Kerberos service principal (service, webauth, etc)
  instance        Kerberos instance
  keytab          keytab location for K5
  ticket_file     Kerberos ticket file location (/tmp/directory.UID.tkt)
  basedn          Search base DNS (dc=stanford, dc=edu)
  scope           Search scope (one,base,sub) default: sub
  weights         Set request control to display attribute weights
  error           Error text
  error_type      Error type - "System" or "Search"
  debug           Defaults to 0.  If set then debugging messages are printed.

B<Stanford::Directory> now supports multiple authentication types.
SASL mechanisms and anonymous binding are available.  If you supply a
I<keytab> and a I<service>, B<Stanford::Directory> will manage the
kerberos tgt for you.  You can also specify where you want the
kerberos ticket stored by using I<ticket_file>.

I<error> and I<error_type> are set whenever an error occurs.

=head1 EXAMPLE for Kerberos

  use Stanford::Directory;

  $dir = new Stanford::Directory;
  $dir->set(ldap_server => 'ldap.stanford.edu',
            mechanism   => 'GSSAPI',
            start_tls   => 1,
            basedn      => 'cn=people, dc=stanford, dc=edu');

  $s = shift or die "usage: $0 <string>\n";
  $filter = "(sn=$s)";

  @entries = $dir->ldap_query($filter, 'displayname');
  if (!@entries) {
      if ($dir->error) {
          print "\n" . $dir->error_type .
                " Error:\n\n   " . $dir->error . "\n\n";
      } else {
          print "No match.\n";
      }
      exit;
  }

  print "People with last names like '$s':\n\n";
  for my $ent (@entries) {
    printf "  %s\n", $ent->displayname;
  }

=head1 EXAMPLE for Kerberos and an array of filters

  use Stanford::Directory;

  $dir = new Stanford::Directory;
  $dir->set(ldap_server => 'ldap.stanford.edu',
            mechanism   => 'GSSAPI',
            start_tls   => 1,
            basedn      => 'cn=people, dc=stanford, dc=edu');

  $s = shift or die "usage: $0 <string>\n";
  @filter = ("sn=$s", "cn=$s");

  %entries = $dir->ldap_query(\@filter, 'displayname');
  if (!%entries) {
      if ($dir->error) {
          print "\n" . $dir->error_type .
                " Error:\n\n   " . $dir->error . "\n\n";
      } else {
          print "No match.\n";
      }
      exit;
  }

  print "People with last names like '$s':\n\n";
  for $key (keys %entries) {
      for my $ent (@{$entries{$key}}) {
          print '  ', $ent->displayname, "\n";
      }
   }

=head1 SEE ALSO

Net::LDAP, Stanford::Directory::Entry, and example scripts provided
in the I<examples> directory.

=head1 AUTHORS

Rob Riepel, riepel@stanford.edu

Quanah Gibson-Mount, quanah@stanford.edu

Bill MacAllister, whm@stanford.edu

Adam H. Lewenberg, adamhl@stanford.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2000-2013, 2020 Board of Trustee, Leland Stanford Jr. University

This module is free software; you may redistribute it and/or modify it
under the same terms as Perl itself.

=cut
